/*
 *	mpegEncode.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/mpegCodec.h>
#include <sp/mpegEncode.h>

#ifdef USE_LAME
#if defined(_WIN32) && !defined(__CYGWIN32__)
#include <lame.h>
#else
#include <lame/lame.h>
#endif
#else
#include <bladeenc/codec.h>
#endif

struct _mpegEncode
{
#ifdef USE_LAME
    lame_global_flags *gfp;
    short buffer[2][1152];
    char filename[SP_MAX_PATHNAME];
#else
    CodecInitIn cinit_in;
#endif
    int init_flag;
};

mpegEncode mpegOpenEncoder(void)
{
    mpegEncode encode;
    
    encode = xalloc(1, struct _mpegEncode);
    encode->init_flag = 0;
#ifdef USE_LAME
#ifdef LAME_OLD
    encode->gfp = xalloc(1, lame_global_flags);
    lame_init(encode->gfp);
#else
    encode->gfp = lame_init();
#endif
#endif

    return encode;
}

int mpegSetEncodeOption(mpegEncode encode, mpegEncodeOption *option)
{
    if (encode == NULL || option == NULL) return 0;
    
#ifdef USE_LAME
#ifdef LAME_OLD
    if (option->option_mask & MPEG_ENC_OPTION_QUALITY) {
	encode->gfp->quality = option->quality;
    }
    if (option->option_mask & MPEG_ENC_OPTION_USE_LOWPASS) {
	if (option->use_lowpass) {
	    encode->gfp->lowpassfreq = 0;
	} else {
	    encode->gfp->lowpassfreq = -1;
	}
    }
    if (option->option_mask & MPEG_ENC_OPTION_USE_VBR) {
	if (option->filename != NULL) {
	    encode->gfp->VBR = option->use_vbr;
	    spStrCopy(encode->filename, sizeof(encode->filename), option->filename);
	    encode->gfp->outPath = encode->filename;
	}
    }
    if (option->option_mask & MPEG_ENC_OPTION_VBR_QUALITY) {
	/*encode->gfp->VBR_q = option->vbr_quality;*/
	encode->gfp->VBR_q = MAX(option->vbr_quality, 1);	/* Quality `0' is not supported yet. */
    }
    if (option->option_mask & MPEG_ENC_OPTION_VBR_MIN_BITRATE) {
	encode->gfp->VBR_min_bitrate_kbps = option->vbr_min_bitrate;
    }
    if (option->option_mask & MPEG_ENC_OPTION_VBR_MAX_BITRATE) {
	encode->gfp->VBR_max_bitrate_kbps = option->vbr_max_bitrate;
    }
#else
    if (option->option_mask & MPEG_ENC_OPTION_QUALITY) {
	lame_set_quality(encode->gfp, option->quality);
    }
    if (option->option_mask & MPEG_ENC_OPTION_USE_LOWPASS) {
	if (option->use_lowpass) {
	    lame_set_lowpassfreq(encode->gfp, 0);
	} else {
	    lame_set_lowpassfreq(encode->gfp, -1);
	}
    }
    if (option->option_mask & MPEG_ENC_OPTION_USE_VBR) {
	if (option->filename != NULL) {
	    spStrCopy(encode->filename, sizeof(encode->filename), option->filename);
	    if (option->use_vbr) {
		lame_set_VBR(encode->gfp, vbr_default);
		lame_set_bWriteVbrTag(encode->gfp, 1);
	    } else {
		lame_set_VBR(encode->gfp, vbr_off);
		lame_set_bWriteVbrTag(encode->gfp, 0);
	    } 
	}
    }
    if (option->option_mask & MPEG_ENC_OPTION_VBR_QUALITY) {
	lame_set_VBR_q(encode->gfp, MAX(option->vbr_quality, 0));
    }
    if (option->option_mask & MPEG_ENC_OPTION_VBR_MIN_BITRATE) {
	lame_set_VBR_min_bitrate_kbps(encode->gfp, option->vbr_min_bitrate);
    }
    if (option->option_mask & MPEG_ENC_OPTION_VBR_MAX_BITRATE) {
	lame_set_VBR_max_bitrate_kbps(encode->gfp, option->vbr_max_bitrate);
    }
#endif
#endif
    
    return 1;
}

int mpegGetEncodeOption(mpegEncode encode, mpegEncodeOption *option)
{
    if (encode == NULL || option == NULL) return 0;
    
#ifdef USE_LAME
    option->option_mask = MPEG_ENC_OPTION_QUALITY | MPEG_ENC_OPTION_USE_LOWPASS
	| MPEG_ENC_OPTION_USE_VBR | MPEG_ENC_OPTION_VBR_QUALITY
        | MPEG_ENC_OPTION_VBR_MIN_BITRATE | MPEG_ENC_OPTION_VBR_MAX_BITRATE;
#ifdef LAME_OLD
    option->quality = encode->gfp->quality;
    if (encode->gfp->lowpassfreq >= 0) {
	option->use_lowpass = 1;
    } else {
	option->use_lowpass = 0;
    }
    option->use_vbr = encode->gfp->VBR;
    option->vbr_quality = encode->gfp->VBR_q;
    option->vbr_min_bitrate = encode->gfp->VBR_min_bitrate_kbps;
    option->vbr_max_bitrate = encode->gfp->VBR_max_bitrate_kbps;
#else
    option->quality = lame_get_quality(encode->gfp);
    if (lame_get_lowpassfreq(encode->gfp) >= 0) {
	option->use_lowpass = 1;
    } else {
	option->use_lowpass = 0;
    }
    option->use_vbr = (lame_get_VBR(encode->gfp) > 0 ? 1 : 0);
    option->vbr_quality = lame_get_VBR_q(encode->gfp);
    option->vbr_min_bitrate = lame_get_VBR_min_bitrate_kbps(encode->gfp);
    option->vbr_max_bitrate = lame_get_VBR_max_bitrate_kbps(encode->gfp);
#endif
#else
    option->option_mask = 0;
#endif
    
    return 1;
}

int mpegGetEncodeInfo(mpegEncode encode, mpegHeader *header, mpegCodecInfo *info)
{
    if (encode == NULL || header == NULL || info == NULL) return 0;

#ifdef USE_LAME
#ifdef LAME_OLD
    encode->gfp->gtkflag = 0;
    encode->gfp->silent = 1;
    encode->gfp->brate = header->bit_rate / 1000;

    header->layer = MPEG_LAYER3_ID;
    if (header->samp_rate < 32000) {
	header->version = MPEG2_ID;
	encode->gfp->brate = MAX(encode->gfp->brate, 56);
	encode->gfp->brate = MIN(encode->gfp->brate, 160);
	encode->gfp->VBR_min_bitrate_kbps = MAX(encode->gfp->VBR_min_bitrate_kbps, 8);
	encode->gfp->VBR_max_bitrate_kbps = MIN(encode->gfp->VBR_max_bitrate_kbps, 160);
    } else {
	header->version = MPEG1_ID;
	encode->gfp->brate = MAX(encode->gfp->brate, 112);
	encode->gfp->brate = MIN(encode->gfp->brate, 320);
	encode->gfp->VBR_min_bitrate_kbps = MAX(encode->gfp->VBR_min_bitrate_kbps, 32);
	encode->gfp->VBR_max_bitrate_kbps = MIN(encode->gfp->VBR_max_bitrate_kbps, 320);
    }
    encode->gfp->VBR_min_bitrate_kbps = MIN(encode->gfp->VBR_min_bitrate_kbps,
					  encode->gfp->VBR_max_bitrate_kbps);
    
    if (header->channel_mode == 3) {
	encode->gfp->num_channels = 1;
    } else {
	encode->gfp->num_channels = 2;
    }
    encode->gfp->in_samplerate = header->samp_rate;
    encode->gfp->mode = header->channel_mode;	/* 0=stereo, 1=jstereo, 3=mono */
    
    encode->gfp->copyright = header->copyright;
    encode->gfp->original = header->original;
    encode->gfp->error_protection = (header->crc_disable == 0);
    encode->gfp->padding_type = header->padding;
    encode->gfp->extension = header->private_bit;
    
    lame_init_params(encode->gfp);

    if (encode->gfp->mode == 3) {	/* mono */
	info->num_channel = 1;
    } else {
	info->num_channel = 2;
    }
    info->samp_rate = encode->gfp->out_samplerate;
    info->bit_rate = encode->gfp->brate * 1000;
#else
    lame_set_brate(encode->gfp, header->bit_rate / 1000);

    header->layer = MPEG_LAYER3_ID;
    if (header->samp_rate < 32000) {
	header->version = MPEG2_ID;
	lame_set_brate(encode->gfp, MAX(lame_get_brate(encode->gfp), 56));
	lame_set_brate(encode->gfp, MIN(lame_get_brate(encode->gfp), 160));
	lame_set_VBR_min_bitrate_kbps(encode->gfp,
				      MAX(lame_get_VBR_min_bitrate_kbps(encode->gfp), 8));
	lame_set_VBR_max_bitrate_kbps(encode->gfp,
				      MIN(lame_get_VBR_max_bitrate_kbps(encode->gfp), 160));
    } else {
	header->version = MPEG1_ID;
	lame_set_brate(encode->gfp, MAX(lame_get_brate(encode->gfp), 112));
	lame_set_brate(encode->gfp, MIN(lame_get_brate(encode->gfp), 320));
	lame_set_VBR_min_bitrate_kbps(encode->gfp,
				      MAX(lame_get_VBR_min_bitrate_kbps(encode->gfp), 32));
	lame_set_VBR_max_bitrate_kbps(encode->gfp,
				      MIN(lame_get_VBR_max_bitrate_kbps(encode->gfp), 320));
    }
    lame_set_VBR_min_bitrate_kbps(encode->gfp,
				  MIN(lame_get_VBR_min_bitrate_kbps(encode->gfp),
				      lame_get_VBR_max_bitrate_kbps(encode->gfp)));
    
    if (header->channel_mode == 3) {
	lame_set_num_channels(encode->gfp, 1);
    } else {
	lame_set_num_channels(encode->gfp, 2);
    }
    lame_set_in_samplerate(encode->gfp, header->samp_rate);
    lame_set_mode(encode->gfp, header->channel_mode);	/* 0=stereo, 1=jstereo, 3=mono */
    
    lame_set_copyright(encode->gfp, header->copyright);
    lame_set_original(encode->gfp, header->original);
    lame_set_error_protection(encode->gfp, (header->crc_disable == 0));
#ifdef LAME_OLD
    lame_set_padding_type(encode->gfp, header->padding);
#endif
    lame_set_extension(encode->gfp, header->private_bit);
    
    lame_init_params(encode->gfp);

    if (lame_get_mode(encode->gfp) == MONO) {	/* mono */
	info->num_channel = 1;
    } else {
	info->num_channel = 2;
    }
    info->samp_rate = lame_get_out_samplerate(encode->gfp);
    info->bit_rate = lame_get_brate(encode->gfp) * 1000;
#endif
    info->max_frame_size = 1152 * 4;
    info->max_output_size = LAME_MAXMP3BUFFER;
    
#else
    if (header->channel_mode != MPEG_CHANNEL_SINGLE_CHANNEL_ID) {
	header->channel_mode = MPEG_CHANNEL_STEREO_ID;	/* stereo only */
    }
    
    if (header->samp_rate != 44100 && header->samp_rate != 48000
	&& header->samp_rate != 32000) {
	spDebug(1, NULL, "Sampling rate must be 32, 44.1 or 48 kHz.\n");
	return 0;
    }
    header->version = MPEG1_ID;
    header->layer = MPEG_LAYER3_ID;
	
    encode->cinit_in.frequency = header->samp_rate;
    encode->cinit_in.mode = header->channel_mode;
    encode->cinit_in.bitrate = header->bit_rate / 1000;
    encode->cinit_in.emphasis = header->emphasis;
    encode->cinit_in.fPrivate = header->private_bit;
    encode->cinit_in.fCRC = header->crc_disable;
    encode->cinit_in.fCopyright = header->copyright;
    encode->cinit_in.fOriginal = header->original;
	
    {
	CodecInitOut *cinit_out;
	
	cinit_out = codecInit(&encode->cinit_in);
	info->max_frame_size = (long)cinit_out->nSamples * 2;
	info->max_output_size = (long)cinit_out->bufferSize;
	spDebug(10, "mpegGetEncodeInfo", "cinit_out->nSamples = %d, cinit_out->bufferSize = %d\n",
		cinit_out->nSamples, cinit_out->bufferSize);
    }
    
    if (encode->cinit_in.mode == 3) {
	info->num_channel = 1;
    } else {
	info->num_channel = 2;
    }
    info->samp_rate = encode->cinit_in.frequency;
    info->bit_rate = encode->cinit_in.bitrate * 1000;
#endif
    info->samp_bit = 16;
    info->data_offset = 0;

    spDebug(10, "mpegGetEncodeInfo", "max_frame_size = %ld, max_output_size = %ld\n",
	    info->max_frame_size, info->max_output_size);
    
    encode->init_flag = 1;
	
    return 1;
}

long mpegEncodeFrame(mpegEncode encode, mpegHeader *header,
		     char *input_buf, long *buf_size, char *output_buf)
{
    long nsample;
    long encoded_size;
    
    if (encode == NULL || input_buf == NULL || buf_size == NULL
	|| *buf_size <= 0 || output_buf == NULL) return -1;

    nsample = *buf_size / 2;
    spDebug(80, "mpegEncodeFrame", "nsample = %ld\n", nsample);
    
#ifdef USE_LAME
    {
	long k;
	short *sbuf = (short *)input_buf;
	
	if (
#ifdef LAME_OLD
	    encode->gfp->mode == 3
#else
	    lame_get_mode(encode->gfp) == MONO
#endif
	    ) { /* mono */
	    encoded_size = lame_encode_buffer(encode->gfp,
					      sbuf, NULL,
					      nsample, (unsigned char *)output_buf, LAME_MAXMP3BUFFER);
	} else {
	    for (k = 0; k < nsample; k++) {
		if (k % 2 == 0) {
		    encode->buffer[0][k/2] = sbuf[k];
		} else {
		    encode->buffer[1][(k-1)/2] = sbuf[k];
		}
	    }
	    nsample /= 2;
	    spDebug(80, "mpegEncodeFrame", "encode stereo data: nsample = %ld\n", nsample);
	    encoded_size = lame_encode_buffer(encode->gfp,
					      encode->buffer[0], encode->buffer[1],
					      nsample, (unsigned char *)output_buf, LAME_MAXMP3BUFFER);
	}
    }
#else
    encoded_size = codecEncodeChunk(nsample, (short *)input_buf, output_buf);
#endif
    spDebug(10, "mpegEncodeFrame", "encoded_size = %ld\n", encoded_size);
    
    return encoded_size;
}

long mpegEncodeLastFrame(mpegEncode encode, mpegHeader *header, char *output_buf)
{
    long encoded_size;
    
    if (encode == NULL || output_buf == NULL) return -1;
    
#ifdef USE_LAME
#ifdef LAME_OLD
    encoded_size = lame_encode_finish(encode->gfp,
				      (unsigned char *)output_buf, LAME_MAXMP3BUFFER);
#else
    encoded_size = lame_encode_flush(encode->gfp,
				     (unsigned char *)output_buf, LAME_MAXMP3BUFFER);
#endif
#else
    encoded_size = codecExit(output_buf);
#endif
    spDebug(10, "mpegEncodeLastFrame", "encoded_size = %ld\n", encoded_size);
    
    return encoded_size;
}

/* You must call this after encoding if you use VBR */
int mpegEncodeTag(mpegEncode encode)
{
    if (encode == NULL) return 0;
    
#ifdef USE_LAME
#ifdef LAME_OLD
    lame_mp3_tags(encode->gfp);
#else
    if (lame_get_bWriteVbrTag(encode->gfp) && !strnone(encode->filename)) {
	FILE *fp;

	if ((fp = fopen(encode->filename, "rb+")) != NULL) {
	    lame_mp3_tags_fid(encode->gfp, fp);
	    fclose(fp);
	}
    }
#endif
#endif
    
    return 1;
}

int mpegEncodeReset(mpegEncode encode)
{
    if (encode == NULL) return 0;

#ifdef USE_LAME
    lame_init_params(encode->gfp);
#else
    codecInit(&encode->cinit_in);
#endif
    
    return 1;
}

int mpegCloseEncoder(mpegEncode encode)
{
    if (encode == NULL) return 0;

#ifdef USE_LAME
#ifdef LAME_OLD
    if (encode->gfp != NULL) xfree(encode->gfp);
#else
    lame_close(encode->gfp);
#endif
#endif
    
    xfree(encode);
    
    return 1;
}
